@@ -1,7 +1,7 @@
-/* $Id: subnetcalc.cc 1023 2011-05-03 17:07:41Z dreibh $
+/* $Id: subnetcalc.cc 1966 2015-03-01 10:42:18Z dreibh $
  *
  * IPv4/IPv6 Subnet Calculator
- * Copyright (C) 2002-2010 by Thomas Dreibholz
+ * Copyright (C) 2002-2015 by Thomas Dreibholz
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -30,8 +30,10 @@
 #include <locale.h>
 #include <netdb.h>
 #include <iostream>
+#include <cstring>
 #ifdef HAVE_GEOIP
 #include <GeoIP.h>
+#include <GeoIPCity.h>
 #endif
 
 #include "tools.h"
@@ -93,8 +95,7 @@ int readPrefix(const char*           parameter,
                const sockaddr_union& forAddress,
                sockaddr_union&       netmask)
 {
-   bool isNumber = false;
-   for(int i = 0;i < strlen(parameter);i++) {
+   for(size_t i = 0;i < strlen(parameter);i++) {
       if(!isdigit(parameter[i])) {
          return(-1);
       }
@@ -183,18 +184,21 @@ void generateUniqueLocal(sockaddr_union& address,
 void printAddressBinary(std::ostream&         os,
                         const sockaddr_union& address,
                         const unsigned int    prefix,
-                        const char*           indent = "")
+                        const bool            colourMode = true,
+                        const char*           indent     = "")
 {
    if(address.sa.sa_family == AF_INET) {
       os << indent;
       uint32_t a = ntohl(getIPv4Address(address));
       for(int i = 31;i >= 0;i--) {
          const uint32_t v = (uint32_t)1 << i;
-         if(32 - i > (int)prefix) {   // Colourize output
-            os << "\x1b[33m";
-         }
-         else {
-            os << "\x1b[34m";
+         if(colourMode) {
+            if(32 - i > (int)prefix) {   // Colourize output
+               os << "\x1b[33m";
+            }
+            else {
+               os << "\x1b[34m";
+            }
          }
          if(a >= v) {
             os << "1";
@@ -203,7 +207,9 @@ void printAddressBinary(std::ostream&         os,
          else {
             os << "0";
          }
-         os << "\x1b[0m";   // Turn off colour printing
+         if(colourMode) {
+            os << "\x1b[0m";   // Turn off colour printing
+         }
          if( ((i % 8) == 0) && (i > 0) ) {
             os << " . ";
          }
@@ -214,17 +220,19 @@ void printAddressBinary(std::ostream&         os,
       int      p           = 0;
       in6_addr ipv6Address = getIPv6Address(address);
       for(int j = 0;j < 8;j++) {
-         uint16_t a = ntohs(((const uint16_t*)&ipv6Address.s6_addr)[j]);
+         uint16_t a = ntohs(ipv6Address.s6_addr16[j]);
          char str[16];
          snprintf((char*)&str, sizeof(str), "%04x", a);
          os << indent << str << " = ";
          for(int i = 15;i >= 0;i--) {
             const uint32_t v = (uint32_t)1 << i;
-            if(p < (int)prefix) {   // Colourize output
-               os << "\x1b[33m";
-            }
-            else {
-               os << "\x1b[34m";
+            if(colourMode) {
+               if(p < (int)prefix) {   // Colourize output
+                  os << "\x1b[33m";
+               }
+               else {
+                  os << "\x1b[34m";
+               }
             }
             if(a >= v) {
                os << "1";
@@ -233,7 +241,9 @@ void printAddressBinary(std::ostream&         os,
             else {
                os << "0";
             }
-            os << "\x1b[0m";   // Turn off colour printing
+            if(colourMode) {
+               os << "\x1b[0m";   // Turn off colour printing
+            }
             if( ((i % 8) == 0) && (i > 0) ) {
                os << " ";
             }
@@ -405,35 +415,37 @@ int operator==(const sockaddr_union& a1, const sockaddr_union& a2)
 // ###### Print IPv6 unicast properties of given address ####################
 void printUnicastProperties(std::ostream&   os,
                             const in6_addr& ipv6address,
+                            const bool      colourMode  = true,
                             const bool      hasSubnetID = true,
                             const bool      hasGlobalID = false)
 {
    // ====== Global ID ======================================================
    if(hasGlobalID) {
-      char           globalIDString[16];
-      const uint16_t globalID = ntohs(((const uint16_t*)&ipv6address.s6_addr)[3]);
+      char globalIDString[16];
       snprintf((char*)&globalIDString, sizeof(globalIDString), "%02x%04x%04x",
-               ipv6address.s6_addr[1],
-               ntohs(((const uint16_t*)&ipv6address.s6_addr)[1]),
-               ntohs(((const uint16_t*)&ipv6address.s6_addr)[2]));
+               ntohs(ipv6address.s6_addr16[0]) & 0xff,
+               ntohs(ipv6address.s6_addr16[1]),
+               ntohs(ipv6address.s6_addr16[2]));
       os << "      + Global ID    = " << globalIDString << std::endl;
    }
 
    // ====== Subnet ID ======================================================
    if(hasSubnetID) {
       char           subnetIDString[16];
-      const uint16_t subnetID = ntohs(((const uint16_t*)&ipv6address.s6_addr)[3]);
+      const uint16_t subnetID = ntohs(ipv6address.s6_addr16[3]);
       snprintf((char*)&subnetIDString, sizeof(subnetIDString), "%04x", subnetID);
       os << "      + Subnet ID    = " << subnetIDString << std::endl;
    }
 
    // ====== Interface ID ===================================================
    char           interfaceIDString[128];
-   const uint16_t interfaceID[4] = { ntohs(((const uint16_t*)&ipv6address.s6_addr)[4]),
-                                     ntohs(((const uint16_t*)&ipv6address.s6_addr)[5]),
-                                     ntohs(((const uint16_t*)&ipv6address.s6_addr)[6]),
-                                     ntohs(((const uint16_t*)&ipv6address.s6_addr)[7]) };
-   snprintf((char*)&interfaceIDString, sizeof(interfaceIDString), "\x1b[36m%04x:%02x\x1b[37m%02x:%02x\x1b[38m%02x:%04x\x1b[0m",
+   const uint16_t interfaceID[4] = { ntohs(ipv6address.s6_addr16[4]),
+                                     ntohs(ipv6address.s6_addr16[5]),
+                                     ntohs(ipv6address.s6_addr16[6]),
+                                     ntohs(ipv6address.s6_addr16[7]) };
+   snprintf((char*)&interfaceIDString, sizeof(interfaceIDString),
+            ((colourMode == true) ? "\x1b[36m%04x:%02x\x1b[37m%02x:%02x\x1b[38m%02x:%04x\x1b[0m" :
+                                    "%04x:%02x%02x:%02x%02x:%04x"),
             interfaceID[0],
             (interfaceID[1] & 0xff00) >> 8, (interfaceID[1] & 0x00ff),
             (interfaceID[2] & 0xff00) >> 8, (interfaceID[2] & 0x00ff),
@@ -442,7 +454,9 @@ void printUnicastProperties(std::ostream&   os,
 
    if( ((interfaceID[1] & 0x00ff) == 0x00ff) &&
        ((interfaceID[2] & 0xff00) == 0xfe00) ) {
-      snprintf((char*)&interfaceIDString, sizeof(interfaceIDString), "\x1b[36m%02x:%02x:%02x\x1b[0m:\x1b[38m%02x:%02x:%02x\x1b[0m",
+      snprintf((char*)&interfaceIDString, sizeof(interfaceIDString),
+               ((colourMode == true) ? "\x1b[36m%02x:%02x:%02x\x1b[0m:\x1b[38m%02x:%02x:%02x\x1b[0m" :
+                                       "%02x:%02x:%02x:%02x:%02x:%02x"),
                ipv6address.s6_addr[8] ^ 0x02,
                ipv6address.s6_addr[9],
                ipv6address.s6_addr[10],
@@ -455,9 +469,10 @@ void printUnicastProperties(std::ostream&   os,
    // ====== Solicited Node Multicast Address ===============================
    char snmcAddressString[32];
    snprintf((char*)&snmcAddressString, sizeof(snmcAddressString),
-            "\x1b[32mff02::1:ff\x1b[38m%02x:%04x\x1b[0m",
-            ntohs(((const uint16_t*)&ipv6address.s6_addr)[6]) & 0xff,
-            ntohs(((const uint16_t*)&ipv6address.s6_addr)[7]));
+            ((colourMode == true) ? "\x1b[32mff02::1:ff\x1b[38m%02x:%04x\x1b[0m" :
+                                    "ff02::1:ff%02x:%04x"),
+            ntohs(ipv6address.s6_addr16[6]) & 0xff,
+            ntohs(ipv6address.s6_addr16[7]));
    os << "      + Sol. Node MC = " << snmcAddressString << std::endl;
 }
 
@@ -468,17 +483,9 @@ void printAddressProperties(std::ostream&         os,
                             const sockaddr_union& netmask,
                             const unsigned int    prefix,
                             const sockaddr_union& network,
-                            const sockaddr_union& broadcast)
+                            const sockaddr_union& broadcast,
+                            const bool            colourMode)
 {
-   in_addr_t ipv4address;
-   in6_addr  ipv6address;
-   if(isIPv4(address)) {
-      ipv4address = ntohl(getIPv4Address(address));
-   }
-   else {
-      ipv6address = getIPv6Address(address);
-   }
-
    // ====== Common properties ==============================================
    os << "Properties    =" << std::endl;
    if(isMulticast(address)) {
@@ -499,8 +506,10 @@ void printAddressProperties(std::ostream&         os,
 
    // ====== IPv4 properties ================================================
    if(isIPv4(address)) {
-      const unsigned int a = ipv4address >> 24;
-      const unsigned int b = (ipv4address & 0x00ff0000) >> 16;
+      const in_addr_t    ipv4address = ntohl(getIPv4Address(address));
+      const unsigned int a           = ipv4address >> 24;
+      const unsigned int b           = (ipv4address & 0x00ff0000) >> 16;
+
       if(IN_CLASSA(ipv4address)) {
          os << "   - Class A" << std::endl;
          if(ipv4address == INADDR_LOOPBACK) {
@@ -568,8 +577,9 @@ void printAddressProperties(std::ostream&         os,
 
    // ====== IPv6 properties ================================================
    else {
-      const uint16_t a = ntohs(((const uint16_t*)&ipv6address.s6_addr)[0]);
-      const uint16_t b = ntohs(((const uint16_t*)&ipv6address.s6_addr)[1]);
+      const in6_addr ipv6address = getIPv6Address(address);
+      const uint16_t a           = ntohs(ipv6address.s6_addr16[0]);
+      const uint16_t b           = ntohs(ipv6address.s6_addr16[1]);
 
       // ------ Special addresses -------------------------------------------
       if(IN6_IS_ADDR_LOOPBACK(&ipv6address)) {
@@ -635,13 +645,13 @@ void printAddressProperties(std::ostream&         os,
          // ------ Solicited node multicast address -------------------------
          // FF02::1:FF00:0/104
          if((a == 0xff02) &&
-            (ntohs(((const uint16_t*)&ipv6address.s6_addr)[5]) == 0x0001) &&
-            (ntohs(((const uint16_t*)&ipv6address.s6_addr)[6]) & 0xff00) == 0xff00) {
+            (ntohs(ipv6address.s6_addr16[5]) == 0x0001) &&
+            (ntohs(ipv6address.s6_addr16[6]) & 0xff00) == 0xff00) {
             char nodeAddressString[64];
             snprintf((char*)&nodeAddressString, sizeof(nodeAddressString),
                      "xxxx:xxxx:xxxx:xxxx:xxxx:xxxx:xx%02x:%04x",
-                     ntohs(((const uint16_t*)&ipv6address.s6_addr)[6]) >> 8,
-                     ntohs(((const uint16_t*)&ipv6address.s6_addr)[7]));
+                     ntohs(ipv6address.s6_addr16[6]) >> 8,
+                     ntohs(ipv6address.s6_addr16[7]));
             os << "      + Address is solicited node multicast address for "
                << nodeAddressString << std::endl;
          }
@@ -650,13 +660,13 @@ void printAddressProperties(std::ostream&         os,
       // ------ Link-local Unicast ------------------------------------------
       else if(IN6_IS_ADDR_LINKLOCAL(&ipv6address)) {
          os << "   - Link-Local Unicast Properties:" << std::endl;
-         printUnicastProperties(std::cout, ipv6address, false, false);
+         printUnicastProperties(std::cout, ipv6address, colourMode, false, false);
       }
 
       // ------ Site-Local Unicast ------------------------------------------
       else if(IN6_IS_ADDR_SITELOCAL(&ipv6address)) {
          os << "   - Site-Local Unicast Properties:" << std::endl;
-         printUnicastProperties(std::cout, ipv6address, true, false);
+         printUnicastProperties(std::cout, ipv6address, colourMode, true, false);
       }
 
       // ------ Unique Local Unicast ----------------------------------------
@@ -668,13 +678,26 @@ void printAddressProperties(std::ostream&         os,
          else {
             os << "      + Assigned by global instance" << std::endl;
          }
-         printUnicastProperties(std::cout, ipv6address, true, true);
+         printUnicastProperties(std::cout, ipv6address, colourMode, true, true);
       }
 
       // ------ Global Unicast ----------------------------------------------
       else if((a & 0xe000) == 0x2000) {
          os << "   - Global Unicast Properties:" << std::endl;
-         printUnicastProperties(std::cout, ipv6address, false, false);
+         printUnicastProperties(std::cout, ipv6address, colourMode, false, false);
+         
+         // ------ 6to4 Address ---------------------------------------------
+         if((a & 0x2002) == 0x2002) {
+            os << "      + 6to4 address = ";
+            sockaddr_union sixToFour;
+            sixToFour.sa.sa_family       = AF_INET;
+            const uint32_t u = ntohs(((const uint16_t*)&ipv6address.s6_addr)[1]);
+            const uint32_t l = ntohs(((const uint16_t*)&ipv6address.s6_addr)[2]);
+            sixToFour.in.sin_addr.s_addr = htonl((u << 16) | l);
+            printAddress(std::cerr, &sixToFour.sa, false);
+            os << std::endl;
+         }
+
       }
    }
 }
@@ -684,6 +707,7 @@ void printAddressProperties(std::ostream&         os,
 // ###### Main program ######################################################
 int main(int argc, char** argv)
 {
+   bool           colourMode      = true;
    bool           noReverseLookup = false;
    int            options;
    int            prefix;
@@ -705,7 +729,7 @@ int main(int argc, char** argv)
    // ====== Print usage ====================================================
    if(argc < 2) {
       printf("Usage: %s [Address{/{Netmask|Prefix}}] "
-             "{Netmask|Prefix} {-n} {-uniquelocal|-uniquelocalhq}\n", argv[0]);
+             "{Netmask|Prefix} {-n} {-uniquelocal|-uniquelocalhq} {-nocolour|-nocolor}\n", argv[0]);
       exit(1);
    }
 
@@ -783,6 +807,9 @@ int main(int argc, char** argv)
       else if(strcmp(argv[i], "-uniquelocalhq") == 0) {
          generateUniqueLocal(address, true);
       }
+      else if( (strcmp(argv[i], "-nocolour") == 0) || (strcmp(argv[i], "-nocolor") == 0) ) {
+         colourMode = false;
+      }
       else if(strcmp(argv[i], "-n") == 0) {
          noReverseLookup = true;
       }
@@ -819,7 +846,7 @@ int main(int argc, char** argv)
 
    // ====== Print results ==================================================
    std::cout << "Address       = " << address << std::endl;
-   printAddressBinary(std::cout, address, prefix, "                   ");
+   printAddressBinary(std::cout, address, prefix, colourMode, "                   ");
    std::cout << "Network       = " << network << " / " << prefix << std::endl;
    std::cout << "Netmask       = " << netmask << std::endl;
    if(isIPv4(address)) {
@@ -842,15 +869,29 @@ int main(int argc, char** argv)
 
 
    // ====== Properties =====================================================
-   printAddressProperties(std::cout, address, netmask, prefix, network, broadcast);
+   printAddressProperties(std::cout, address, netmask, prefix, network, broadcast, colourMode);
 
 
    // ====== GeoIP ==========================================================
 #ifdef HAVE_GEOIP
+   // libGeoIP prints an error message each time it cannot open a database.
+   // Unfortunately, packaging of these databases is very confusing: some
+   // distributions include some of them, etc.. Just avoid annoying the user
+   // by printing these errors to /dev/null. The error condition of a
+   // non-existing database is handled by subnetcalc anyway.
+   if(freopen("/dev/null", "w", stderr) == NULL) { }
+
    const char* country = NULL;
    const char* code    = NULL;
    if(address.sa.sa_family == AF_INET) {
-      GeoIP* geoIP = GeoIP_open_type(GEOIP_COUNTRY_EDITION, GEOIP_STANDARD);
+      GeoIP* geoIP = GeoIP_open_type(GEOIP_ASNUM_EDITION, GEOIP_STANDARD);
+      if(geoIP) {
+         const char* org = GeoIP_name_by_ipnum(geoIP, ntohl(address.in.sin_addr.s_addr));
+         std::cout << "GeoIP AS Info = "
+                   << ((org != NULL) ? org : "Unknown") << std::endl;
+         GeoIP_delete(geoIP);
+      }
+      geoIP = GeoIP_open_type(GEOIP_COUNTRY_EDITION, GEOIP_STANDARD);
       if(geoIP) {
          country = GeoIP_country_name_by_ipnum(geoIP, ntohl(address.in.sin_addr.s_addr));
          code    = GeoIP_country_code_by_ipnum(geoIP, ntohl(address.in.sin_addr.s_addr));
@@ -859,10 +900,39 @@ int main(int argc, char** argv)
                    << " (" << ((code != NULL) ? code : "??") << ")" << std::endl;
          GeoIP_delete(geoIP);
       }
+      geoIP = GeoIP_open_type(GEOIP_CITY_EDITION_REV1, GEOIP_STANDARD);
+      if(geoIP) {
+         GeoIPRecord* gir = GeoIP_record_by_ipnum(geoIP, ntohl(address.in.sin_addr.s_addr));
+         if(gir != NULL) {
+            const char* timeZone = GeoIP_time_zone_by_country_and_region(
+                                      gir->country_code, gir->region);
+            std::cout << "GeoIP Region  = "
+                      << ((gir->postal_code != NULL) ? gir->postal_code : "")
+                      << ((gir->postal_code != NULL) ? " " : "")
+                      << ((gir->city != NULL) ? gir->city : "Unknown")
+                      << ", " << ((gir->region != NULL) ? gir->region : "Unknown")
+                      << " ("
+                      << fabs(gir->latitude)  << "°" << ((gir->latitude > 0)  ? "N" : "S") << ", "
+                      << fabs(gir->longitude) << "°" << ((gir->longitude > 0) ? "E" : "W")
+                      << ((timeZone != NULL) ? ", " : "")
+                      << ((timeZone != NULL) ? timeZone : "")
+                      << ")"
+                      << std::endl;
+            GeoIPRecord_delete(gir);
+         }
+         GeoIP_delete(geoIP);
+      }
    }
 #ifdef HAVE_GEOIP_IPV6
    else if(address.sa.sa_family == AF_INET6) {
-      GeoIP* geoIP = GeoIP_open_type(GEOIP_COUNTRY_EDITION_V6, GEOIP_STANDARD);
+      GeoIP* geoIP = GeoIP_open_type(GEOIP_ASNUM_EDITION_V6, GEOIP_STANDARD);
+      if(geoIP) {
+         const char* org = GeoIP_name_by_ipnum_v6(geoIP, address.in6.sin6_addr);
+         std::cout << "GeoIP AS Info = "
+                   << ((org != NULL) ? org : "Unknown") << std::endl;
+         GeoIP_delete(geoIP);
+      }
+      geoIP = GeoIP_open_type(GEOIP_COUNTRY_EDITION_V6, GEOIP_STANDARD);
       if(geoIP) {
          country = GeoIP_country_name_by_ipnum_v6(geoIP, address.in6.sin6_addr);
          code    = GeoIP_country_code_by_ipnum_v6(geoIP, address.in6.sin6_addr);
@@ -871,6 +941,28 @@ int main(int argc, char** argv)
                    << " (" << ((code != NULL) ? code : "??") << ")" << std::endl;
          GeoIP_delete(geoIP);
       }
+      geoIP = GeoIP_open_type(GEOIP_CITY_EDITION_REV1_V6, GEOIP_STANDARD);
+      if(geoIP) {
+         GeoIPRecord* gir = GeoIP_record_by_ipnum_v6(geoIP, address.in6.sin6_addr);
+         if(gir != NULL) {
+            const char* timeZone = GeoIP_time_zone_by_country_and_region(
+                                      gir->country_code, gir->region);
+            std::cout << "GeoIP Region  = "
+                      << ((gir->postal_code != NULL) ? gir->postal_code : "")
+                      << ((gir->postal_code != NULL) ? " " : "")
+                      << ((gir->city != NULL) ? gir->city : "Unknown")
+                      << ", " << ((gir->region != NULL) ? gir->region : "Unknown")
+                      << " ("
+                      << fabs(gir->latitude)  << "°" << ((gir->latitude > 0)  ? "N" : "S") << ", "
+                      << fabs(gir->longitude) << "°" << ((gir->longitude > 0) ? "E" : "W")
+                      << ((timeZone != NULL) ? ", " : "")
+                      << ((timeZone != NULL) ? timeZone : "")
+                      << ")"
+                      << std::endl;
+            GeoIPRecord_delete(gir);
+         }
+         GeoIP_delete(geoIP);
+      }
    }
 #else
 #warning This used version of GeoIP does not yet support IPv6!
@@ -894,7 +986,7 @@ int main(int argc, char** argv)
                               NI_NAMEREQD
 #endif
                               );
-      std::cout << "\r\x1b[K";
+      std::cout << ((colourMode == true) ? "\r\x1b[K" : "\r");
       std::cout << "DNS Hostname  = "; std::cout.flush();
       if(error == 0) {
          std::cout << hostname << std::endl;
